/*
 * Copyright 2013 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "JoystickDriver.h"
#include <sstream>
#include <limits>
#include <cstdlib>

JoystickDriver::JoystickDriver(const std::string& name, SDL_Joystick* joy)
:	name(name),
	joy(joy)
{
	for (int i = 0; i < 4; ++i)
	{
		inputs[i].type = JOY_UNKNOWN;
		inputs[i].index = -1;
		inputs[i].state = -1;
	}
}

int JoystickDriver::drive(int carId, const Track& track)
{
	if (joy == 0)
	{
		return 0;
	}
	
	int flags = 0;
	for (int i = 0; i < 4; ++i)
	{
		switch (inputs[i].type)
		{
		case JOY_AXIS:
			{
				Sint16 axis = SDL_JoystickGetAxis(joy, inputs[i].index);
				if (inputs[i].state == 0)
				{
					if (axis > AXIS_THRESHOLD)
					{
						flags |= 1 << i;
					}
				}
				else
				{
					if (axis < -AXIS_THRESHOLD)
					{
						flags |= 1 << i;
					}
				}
			}
			break;
		case JOY_HAT:
			if (SDL_JoystickGetHat(joy, inputs[i].index) & inputs[i].state)
			{
				flags |= 1 << i;
			}
			break;
		case JOY_BUTTON:
			if (SDL_JoystickGetButton(joy, inputs[i].index) != 0)
			{
				flags |= 1 << i;
			}
			break;
		default:
			;
		}
	}
	return flags;
}

bool JoystickDriver::setControl(const SDL_Event& event, int i)
{
	if (joy == 0)
	{
		return false;
	}
	
	int joyIndex = SDL_JoystickIndex(joy);
	switch (event.type)
	{
	case SDL_JOYAXISMOTION:
		if (event.jaxis.which == joyIndex && std::abs(event.jaxis.value) > AXIS_THRESHOLD)
		{
			inputs[i].type = JOY_AXIS;
			inputs[i].index = event.jaxis.axis;
			inputs[i].state = (event.jaxis.value > 0 ? 0 : 1);
			return true;
		}
		break;
	case SDL_JOYHATMOTION:
		if (event.jhat.which == joyIndex && (event.jhat.value & ~SDL_HAT_CENTERED))
		{
			inputs[i].type = JOY_HAT;
			inputs[i].index = event.jhat.hat;
			int bit = 1;
			while (!(event.jhat.value & bit))
			{
				bit <<= 1;
			}
			inputs[i].state = bit;
			return true;
		}
		break;
	case SDL_JOYBUTTONDOWN:
		if (event.jbutton.which == joyIndex)
		{
			inputs[i].type = JOY_BUTTON;
			inputs[i].index = event.jbutton.button;
			inputs[i].state = 0;
			return true;
		}
		break;
	}
	return false;
}

std::string JoystickDriver::getControlName(int i) const
{
	std::ostringstream oss;
	switch (inputs[i].type)
	{
	case JOY_AXIS:
		oss << "axis " << (inputs[i].index + 1) << (inputs[i].state == 0 ? '+' : '-');
		break;
	case JOY_HAT:
		oss << "hat " << (inputs[i].index + 1) << ' ';
		if (inputs[i].state & SDL_HAT_LEFT)
		{
			oss << "left";
		}
		else if (inputs[i].state & SDL_HAT_RIGHT)
		{
			oss << "right";
		}
		else if (inputs[i].state & SDL_HAT_UP)
		{
			oss << "up";
		}
		else if (inputs[i].state & SDL_HAT_DOWN)
		{
			oss << "down";
		}
		break;
	case JOY_BUTTON:
		oss << "button " << (inputs[i].index + 1);
		break;
	default:
		;
	}
	return oss.str();
}

void JoystickDriver::save(std::ostream& out) const
{
	out << name << '\n';
	for (int i = 0; i < 4; ++i)
	{
		out << ' ' << inputs[i].type << ' ' << inputs[i].index << ' ' << inputs[i].state;
	}
	out << '\n';
}

void JoystickDriver::load(std::istream& in)
{
	for (int i = 0; i < 4; ++i)
	{
		in >> inputs[i].type >> inputs[i].index >> inputs[i].state;
	}
	in.ignore(std::numeric_limits<std::streamsize>::max(), '\n');
}
